@@ -25,7 +25,8 @@
 #include "common/util.h"
 
 #include "graphics/macega.h"
-#include "graphics/palette.h"
+#include "graphics/macgamma.h"
+#include "graphics/paletteman.h"
 
 #include "scumm/resource.h"
 #include "scumm/scumm.h"
@@ -33,6 +34,7 @@
 #include "scumm/scumm_v8.h"
 #include "scumm/util.h"
 #include "scumm/charset.h"
+#include "scumm/macgui/macgui.h"
 
 namespace Scumm {
 
@@ -53,7 +55,7 @@ uint16 ScummEngine::get16BitColor(uint8 r, uint8 g, uint8 b) {
 	return _outputPixelFormat.RGBToColor(r, g, b);
 }
 
-void ScummEngine::resetPalette() {
+void ScummEngine::resetPalette(bool isBootUp) {
 	static const byte tableC64Palette[] = {
 #if 1  // VICE-based palette. See bug #4576
 		0x00, 0x00, 0x00,	0xFF, 0xFF, 0xFF,	0x7E, 0x35, 0x2B,	0x6E, 0xB7, 0xC1,
@@ -61,10 +63,10 @@ void ScummEngine::resetPalette() {
 		0x85, 0x53, 0x1C,	0x50, 0x3C, 0x00,	0xB4, 0x6B, 0x61,	0x4A, 0x4A, 0x4A,
 		0x75, 0x75, 0x75,	0xA3, 0xE7, 0x7C,	0x70, 0x64, 0xD6,	0xA3, 0xA3, 0xA3,
 #else
-		0x00, 0x00, 0x00, 	0xFD, 0xFE, 0xFC, 	0xBE, 0x1A, 0x24, 	0x30, 0xE6, 0xC6,
-		0xB4, 0x1A, 0xE2, 	0x1F, 0xD2, 0x1E, 	0x21, 0x1B, 0xAE, 	0xDF, 0xF6, 0x0A,
-		0xB8, 0x41, 0x04, 	0x6A, 0x33, 0x04, 	0xFE, 0x4A, 0x57, 	0x42, 0x45, 0x40,
-		0x70, 0x74, 0x6F, 	0x59, 0xFE, 0x59, 	0x5F, 0x53, 0xFE, 	0xA4, 0xA7, 0xA2,
+		0x00, 0x00, 0x00,	0xFD, 0xFE, 0xFC,	0xBE, 0x1A, 0x24,	0x30, 0xE6, 0xC6,
+		0xB4, 0x1A, 0xE2,	0x1F, 0xD2, 0x1E,	0x21, 0x1B, 0xAE,	0xDF, 0xF6, 0x0A,
+		0xB8, 0x41, 0x04,	0x6A, 0x33, 0x04,	0xFE, 0x4A, 0x57,	0x42, 0x45, 0x40,
+		0x70, 0x74, 0x6F,	0x59, 0xFE, 0x59,	0x5F, 0x53, 0xFE,	0xA4, 0xA7, 0xA2,
 #endif
 		// Use 17 color table for v1 games to allow correct color for inventory and
 		// sentence line. Original games used some kind of dynamic color table
@@ -124,24 +126,24 @@ void ScummEngine::resetPalette() {
 	};
 
 	static const byte tableAmigaPalette[] = {
-		0x00, 0x00, 0x00, 	0x00, 0x00, 0xBB, 	0x00, 0xBB, 0x00, 	0x00, 0xBB, 0xBB,
-		0xBB, 0x00, 0x00, 	0xBB, 0x00, 0xBB, 	0xBB, 0x77, 0x00, 	0xBB, 0xBB, 0xBB,
-		0x77, 0x77, 0x77, 	0x77, 0x77, 0xFF, 	0x00, 0xFF, 0x00, 	0x00, 0xFF, 0xFF,
-		0xFF, 0x88, 0x88, 	0xFF, 0x00, 0xFF, 	0xFF, 0xFF, 0x00, 	0xFF, 0xFF, 0xFF
+		0x00, 0x00, 0x00,	0x00, 0x00, 0xBB,	0x00, 0xBB, 0x00,	0x00, 0xBB, 0xBB,
+		0xBB, 0x00, 0x00,	0xBB, 0x00, 0xBB,	0xBB, 0x77, 0x00,	0xBB, 0xBB, 0xBB,
+		0x77, 0x77, 0x77,	0x77, 0x77, 0xFF,	0x00, 0xFF, 0x00,	0x00, 0xFF, 0xFF,
+		0xFF, 0x88, 0x88,	0xFF, 0x00, 0xFF,	0xFF, 0xFF, 0x00,	0xFF, 0xFF, 0xFF
 	};
 
 	static const byte tableAmigaMIPalette[] = {
-		0x00, 0x00, 0x00, 	0x00, 0x00, 0xAA, 	0x00, 0x88, 0x22, 	0x00, 0x66, 0x77,
-		0xBB, 0x66, 0x66, 	0xAA, 0x22, 0xAA, 	0x88, 0x55, 0x22, 	0x77, 0x77, 0x77,
-		0x33, 0x33, 0x33, 	0x22, 0x55, 0xDD, 	0x22, 0xDD, 0x44, 	0x00, 0xCC, 0xFF,
-		0xFF, 0x99, 0x99, 	0xFF, 0x55, 0xFF, 	0xFF, 0xFF, 0x77, 	0xFF, 0xFF, 0xFF
+		0x00, 0x00, 0x00,	0x00, 0x00, 0xAA,	0x00, 0x88, 0x22,	0x00, 0x66, 0x77,
+		0xBB, 0x66, 0x66,	0xAA, 0x22, 0xAA,	0x88, 0x55, 0x22,	0x77, 0x77, 0x77,
+		0x33, 0x33, 0x33,	0x22, 0x55, 0xDD,	0x22, 0xDD, 0x44,	0x00, 0xCC, 0xFF,
+		0xFF, 0x99, 0x99,	0xFF, 0x55, 0xFF,	0xFF, 0xFF, 0x77,	0xFF, 0xFF, 0xFF
 	};
 
 	static const byte tableEGAPalette[] = {
-		0x00, 0x00, 0x00, 	0x00, 0x00, 0xAA, 	0x00, 0xAA, 0x00, 	0x00, 0xAA, 0xAA,
-		0xAA, 0x00, 0x00, 	0xAA, 0x00, 0xAA, 	0xAA, 0x55, 0x00, 	0xAA, 0xAA, 0xAA,
-		0x55, 0x55, 0x55, 	0x55, 0x55, 0xFF, 	0x55, 0xFF, 0x55, 	0x55, 0xFF, 0xFF,
-		0xFF, 0x55, 0x55, 	0xFF, 0x55, 0xFF, 	0xFF, 0xFF, 0x55, 	0xFF, 0xFF, 0xFF
+		0x00, 0x00, 0x00,	0x00, 0x00, 0xAA,	0x00, 0xAA, 0x00,	0x00, 0xAA, 0xAA,
+		0xAA, 0x00, 0x00,	0xAA, 0x00, 0xAA,	0xAA, 0x55, 0x00,	0xAA, 0xAA, 0xAA,
+		0x55, 0x55, 0x55,	0x55, 0x55, 0xFF,	0x55, 0xFF, 0x55,	0x55, 0xFF, 0xFF,
+		0xFF, 0x55, 0x55,	0xFF, 0x55, 0xFF,	0xFF, 0xFF, 0x55,	0xFF, 0xFF, 0xFF
 	};
 
 	static const byte _cgaColors[4][12] = {
@@ -152,11 +154,11 @@ void ScummEngine::resetPalette() {
 	};
 
 	static const byte tableHercAPalette[] = {
-		0x00, 0x00, 0x00, 	0xAE, 0x69, 0x38
+		0x00, 0x00, 0x00,	0xAE, 0x69, 0x38
 	};
 
 	static const byte tableHercGPalette[] = {
-		0x00, 0x00, 0x00, 	0x00, 0xFF, 0x00
+		0x00, 0x00, 0x00,	0x00, 0xFF, 0x00
 	};
 
 	// Palette based on Apple IIgs Technical Notes: IIgs 2523063 Master Color Values
@@ -223,10 +225,6 @@ void ScummEngine::resetPalette() {
 	} else if (_game.features & GF_16COLOR) {
 		switch (_renderMode) {
 		case Common::kRenderEGA:
-		case Common::kRenderMacintoshBW:
-			// Use EGA palette for MacintoshBW, because that makes
-			// white 0xFFFFFF there. The Mac EGA palette, would
-			// make it 0xFCFCFC.
 			setPaletteFromTable(tableEGAPalette, sizeof(tableEGAPalette) / 3);
 			break;
 
@@ -237,18 +235,19 @@ void ScummEngine::resetPalette() {
 		default:
 			if ((_game.platform == Common::kPlatformAmiga) || (_game.platform == Common::kPlatformAtariST))
 				setPaletteFromTable(tableAmigaPalette, sizeof(tableAmigaPalette) / 3);
-			else if ((_game.id == GID_LOOM || _game.id == GID_INDY3) && _game.platform == Common::kPlatformMacintosh)
-				setPaletteFromTable(Graphics::macEGAPalette, sizeof(Graphics::macEGAPalette) / 3);
 			else
 				setPaletteFromTable(tableEGAPalette, sizeof(tableEGAPalette) / 3);
 		}
-
+		setDirtyColors(0, 255);
 	} else {
 		if ((_game.platform == Common::kPlatformAmiga) && _game.version == 4) {
 			// if rendermode is set to EGA we use the full palette from the resources
 			// else we initialize and then lock down the first 16 colors.
-			if (_renderMode != Common::kRenderEGA)
+			if (_renderMode != Common::kRenderEGA) {
 				setPaletteFromTable(tableAmigaMIPalette, sizeof(tableAmigaMIPalette) / 3);
+			} else {
+				setPaletteFromTable(tableEGAPalette, sizeof(tableEGAPalette) / 3);
+			}
 		} else if (_renderMode == Common::kRenderEGA && _supportsEGADithering) {
 			setPaletteFromTable(tableEGAPalette, sizeof(tableEGAPalette) / 3);
 			_enableEGADithering = true;
@@ -263,7 +262,7 @@ void ScummEngine::resetPalette() {
 			if (_game.id == GID_INDY4 || _game.id == GID_MONKEY2)
 				_townsClearLayerFlag = 0;
 #ifdef USE_RGB_COLOR
-			else if (_game.id == GID_LOOM)
+			else if (_game.id == GID_LOOM || _game.id == GID_MONKEY) // Setting a default for MI1 as well!
 				towns_setTextPaletteFromPtr(tableTownsLoomPalette);
 			else if (_game.version == 3)
 				towns_setTextPaletteFromPtr(tableTownsV3Palette);
@@ -271,9 +270,26 @@ void ScummEngine::resetPalette() {
 
 			_townsScreen->toggleLayers(_townsActiveLayerFlags);
 #endif // DISABLE_TOWNS_DUAL_LAYER_MODE
+		} else if (isBootUp && (_game.version >= 4 && _game.heversion == 0) &&
+			(_game.platform == Common::kPlatformDOS || _game.platform == Common::kPlatformUnknown) &&
+			_renderMode == Common::kRenderDefault) {
+			// VGA games at boot-up have at their disposal whatever mode 13h
+			// is offering at that moment: the default palette.
+			// We just need the first few colors, which are the same as the EGA colors.
+			// (See #15869: "SCUMM: MI1: Message banner can be invisible if palette is all black")
+			setPaletteFromTable(tableEGAPalette, sizeof(tableEGAPalette) / 3);
+
+			if (_game.id == GID_MONKEY_VGA) {
+				for (int i = 0; i < _shadowPaletteSize; i++)
+					_shadowPalette[i] = i;
+			}
 		}
+
 		setDirtyColors(0, 255);
 	}
+
+	if (isBootUp && _game.heversion == 0)
+		updatePalette();
 }
 
 void ScummEngine::setPaletteFromTable(const byte *ptr, int numcolor, int index) {
@@ -415,6 +431,9 @@ void ScummEngine::setPaletteFromPtr(const byte *ptr, int numcolor) {
 	}
 
 	setDirtyColors(firstIndex, numcolor - 1);
+
+	if (_game.id == GID_SAMNMAX && !enhancementEnabled(kEnhMinorBugFixes))
+		VAR(77) = 0;
 }
 
 void ScummEngine::setAmigaPaletteFromPtr(const byte *ptr) {
@@ -474,7 +493,7 @@ void ScummEngine::setV1ColorTable(int renderMode) {
 		{	0x00, 0x0F, 0x08, 0x05, 0x0A, 0x05, 0x01, 0x0D, 0x0A, 0x02, 0x0A, 0x0C, 0x0F, 0x0A, 0x05, 0x0C }
 	};
 
-	int tbl = (_game.platform == Common::kPlatformC64) ? 0 : (_game.id == GID_ZAK ? 1 : 3);
+	int tbl = (_game.platform == Common::kPlatformC64 || _game.platform == Common::kPlatformApple2GS) ? 0 : (_game.id == GID_ZAK ? 1 : 3);
 	if (renderMode == Common::kRenderHercA || renderMode == Common::kRenderHercG || renderMode == Common::kRenderCGA || renderMode == Common::kRenderCGA_BW)
 		++tbl;
 
@@ -581,6 +600,8 @@ void ScummEngine::setDirtyColors(int min, int max) {
 		_palDirtyMin = min;
 	if (_palDirtyMax < max)
 		_palDirtyMax = max;
+
+	_paletteChangedCounter++; // HE99+
 }
 
 void ScummEngine::initCycl(const byte *ptr) {
@@ -590,29 +611,20 @@ void ScummEngine::initCycl(const byte *ptr) {
 	memset(_colorCycle, 0, sizeof(_colorCycle));
 
 	if (_game.features & GF_SMALL_HEADER) {
+		// Code verified from LOOM CD and MI1 VGA disasms, which
+		// are the only executables which contain said code at all
+
 		cycl = _colorCycle;
 		for (j = 0; j < 16; ++j, ++cycl) {
-			uint16 delay = READ_BE_UINT16(ptr);
+			cycl->delay = READ_BE_UINT16(ptr);
 			ptr += 2;
-			byte start = *ptr++;
-			byte end = *ptr++;
-
-			if (!delay || delay == 0x0aaa || start >= end)
-				continue;
 
 			cycl->counter = 0;
-			cycl->delay = 16384 / delay;
-			cycl->flags = 2;
-			// FIXME bug #10854: lava flows up instead of down in the floppy VGA version
-			// of Monkey1 if we don't do this. It's fine in the original interpreter and
-			// in the VGA CD version (which doesn't use GF_SMALL_HEADER). This is maybe
-			// meant to be always 0 for GF_SMALL_HEADER games, but until disasm confirms
-			// or disproves this, we limit this change to the lava cave (where we can see
-			// it's wrong), just in case.
-			if (_game.id == GID_MONKEY_VGA && (_roomResource == 39 || _roomResource == 65))
-				cycl->flags = 0;
-			cycl->start = start;
-			cycl->end = end;
+			cycl->start = *ptr++;
+			cycl->end = *ptr++;
+
+			if (cycl->delay && cycl->delay != 0x0AAA)
+				cycl->counter = cycl->start;
 		}
 	} else {
 		memset(_colorUsedByCycle, 0, sizeof(_colorUsedByCycle));
@@ -734,6 +746,35 @@ void ScummEngine::cyclePalette() {
 	int valueToAdd;
 	int i, j;
 
+	if (_game.features & GF_SMALL_HEADER) {
+		// Code verified from LOOM CD and MI1 VGA disasms, which
+		// are the only executables which contain said code at all
+
+		for (i = 0; i < 16; i++) {
+			if (_colorCycle[i].counter) {
+				_colorCycle[i].counter++;
+				if (_colorCycle[i].counter > _colorCycle[i].end)
+					_colorCycle[i].counter = _colorCycle[i].start;
+
+				byte start = _colorCycle[i].start;
+				byte end = _colorCycle[i].end;
+
+				if (start <= end) {
+					byte cycleVal = _colorCycle[i].counter;
+					for (j = start; j <= end; j++) {
+						_shadowPalette[j] = cycleVal--;
+						if (cycleVal < start)
+							cycleVal = end;
+					}
+				}
+
+				setDirtyColors(_colorCycle[i].start, _colorCycle[i].end);
+				moveMemInPalRes(_colorCycle[i].start, _colorCycle[i].end, 0);
+			}
+		}
+		return;
+	}
+
 #ifndef DISABLE_TOWNS_DUAL_LAYER_MODE
 	if (_game.platform == Common::kPlatformFMTowns && !(_townsPaletteFlags & 1))
 		return;
@@ -989,10 +1030,10 @@ void ScummEngine::setShadowPalette(int redScale, int greenScale, int blueScale,
 byte egaFindBestMatch(int r, int g, int b) {
 	// This is almost like the normal EGA palette, but a bit different
 	static const byte matchPalette[] = {
-		0x00, 0x00, 0x00, 	0x00, 0x00, 0xAB, 	0x00, 0xAB, 0x00, 	0x00, 0xAB, 0xAB,
-		0xAB, 0x00, 0x00, 	0xAB, 0x00, 0xAB, 	0xAB, 0x57, 0x00, 	0xAB, 0xAB, 0xAB,
-		0x57, 0x57, 0x57, 	0x57, 0x57, 0xFF, 	0x57, 0xFF, 0x57, 	0x57, 0xFF, 0xFF,
-		0xFF, 0x57, 0x57, 	0xFF, 0x57, 0xFF, 	0xFF, 0xFF, 0x57, 	0xFF, 0xFF, 0xFF
+		0x00, 0x00, 0x00,	0x00, 0x00, 0xAB,	0x00, 0xAB, 0x00,	0x00, 0xAB, 0xAB,
+		0xAB, 0x00, 0x00,	0xAB, 0x00, 0xAB,	0xAB, 0x57, 0x00,	0xAB, 0xAB, 0xAB,
+		0x57, 0x57, 0x57,	0x57, 0x57, 0xFF,	0x57, 0xFF, 0x57,	0x57, 0xFF, 0xFF,
+		0xFF, 0x57, 0x57,	0xFF, 0x57, 0xFF,	0xFF, 0xFF, 0x57,	0xFF, 0xFF, 0xFF
 	};
 
 	uint32 best = (uint32)-1;
@@ -1154,6 +1195,10 @@ void ScummEngine::darkenPalette(int redScale, int greenScale, int blueScale, int
 			}
 			if (_game.heversion != 70)
 				setDirtyColors(startColor, endColor);
+
+			// Original noir mode behavior
+			if (_game.id == GID_SAMNMAX && !enhancementEnabled(kEnhMinorBugFixes) && VAR(77) == 1)
+				applyGrayscaleToPaletteRange(startColor, endColor);
 		}
 	}
 }
@@ -1358,6 +1403,10 @@ void ScummEngine::copyPalColor(int dst, int src) {
 		_16BitPalette[dst] = get16BitColor(sp[0], sp[1], sp[2]);
 
 	setDirtyColors(dst, dst);
+
+	// Original noir mode behavior
+	if (_game.id == GID_SAMNMAX && !enhancementEnabled(kEnhMinorBugFixes) && VAR(77) == 1)
+		applyGrayscaleToPaletteRange(src, src);
 }
 
 void ScummEngine::setPalColor(int idx, int r, int g, int b) {
@@ -1421,6 +1470,10 @@ void ScummEngine::setPalColor(int idx, int r, int g, int b) {
 		_16BitPalette[idx] = get16BitColor(r, g, b);
 
 	setDirtyColors(idx, idx);
+
+	// Original noir mode behavior
+	if (_game.id == GID_SAMNMAX && !enhancementEnabled(kEnhMinorBugFixes) && VAR(77) == 1)
+		applyGrayscaleToPaletteRange(idx, idx);
 }
 
 void ScummEngine::setCurrentPalette(int palindex) {
@@ -1554,7 +1607,51 @@ void ScummEngine::fetchBlackAndWhite(uint32 &black, uint32 &white, byte *palette
 
 uint32 ScummEngine::findClosestPaletteColor(byte *palette, int paletteLength, byte r, byte g, byte b) {
 	_pl.setPalette(palette, paletteLength);
-	return (uint32)_pl.findBestColor(r, g, b, true);
+	return (uint32)_pl.findBestColor(r, g, b, Graphics::kColorDistanceNaive);
+}
+
+void ScummEngine::applyGrayscaleToPaletteRange(int min, int max) {
+	assertRange(0, min, 256, "ScummEngine::applyGrayscaleToPaletteRange(): min");
+	assertRange(0, max, 256, "ScummEngine::applyGrayscaleToPaletteRange(): min");
+
+	if (min <= max) {
+		int count = max - min + 1;
+		byte *paletteEntry = &_currentPalette[3 * min];
+
+		for (int i = 0; i < count; i++) {
+			int r = paletteEntry[0];
+			int g = paletteEntry[1];
+			int b = paletteEntry[2];
+			byte average = (byte)((r + g + b) / 3);
+
+			// The original checks for colors to be < 126,
+			// but we handle things a little bit differently,
+			// so this is not needed...
+			//
+			// if (r < 126 || g < 126 || b < 126) {
+				paletteEntry[0] = average;
+				paletteEntry[1] = average;
+				paletteEntry[2] = average;
+			// }
+
+			paletteEntry += 3;
+		}
+	}
+
+	setDirtyColors(min, max);
+}
+
+bool ScummEngine::haveToApplyMonkey1PaletteFix() {
+	if (_game.id != GID_MONKEY)
+		return false;
+
+	bool canChangeMonkey1PaletteSlot = _game.platform == Common::kPlatformMacintosh;
+
+	canChangeMonkey1PaletteSlot |= enhancementEnabled(kEnhVisualChanges) &&
+								   (_game.platform != Common::kPlatformSegaCD && _game.platform != Common::kPlatformFMTowns &&
+									!(_game.features & GF_ULTIMATE_TALKIE));
+
+	return canChangeMonkey1PaletteSlot;
 }
 
 void ScummEngine::updatePalette() {
@@ -1564,8 +1661,8 @@ void ScummEngine::updatePalette() {
 	if (_palDirtyMax == -1)
 		return;
 
-	byte palette_colors[3 * 256];
-	byte *p = palette_colors;
+	byte paletteColors[3 * 256];
+	byte *p = paletteColors;
 	int first;
 	int num;
 
@@ -1597,7 +1694,7 @@ void ScummEngine::updatePalette() {
 
 		_system->getPaletteManager()->setPalette(mouseCursorPalette, 252, 3);
 	} else {
-		bool noir_mode = (_game.id == GID_SAMNMAX && readVar(0x8000));
+		bool noirMode = (_game.id == GID_SAMNMAX && readVar(0x8000));
 		int i;
 
 		first = _palDirtyMin;
@@ -1606,19 +1703,14 @@ void ScummEngine::updatePalette() {
 		for (i = _palDirtyMin; i <= _palDirtyMax; i++) {
 			byte *data;
 
-			// In b/w Mac rendering mode, the shadow palette is
-			// handled by the renderer itself. See comment in
-			// mac_drawStripToScreen().
-
-			if (_game.features & GF_SMALL_HEADER && _game.version > 2 && _renderMode != Common::kRenderMacintoshBW)
+			if (_shadowPalRemap)
 				data = _currentPalette + _shadowPalette[i] * 3;
 			else
 				data = _currentPalette + i * 3;
 
-			// Sam & Max film noir mode. Convert the colors to grayscale
-			// before uploading them to the backend.
-
-			if (noir_mode) {
+			// Sam & Max film noir mode - Enhanced version.
+			// Convert the colors to grayscale before uploading them to the backend.
+			if (noirMode && enhancementEnabled(kEnhMinorBugFixes)) {
 				int r, g, b;
 				byte brightness;
 
@@ -1626,7 +1718,7 @@ void ScummEngine::updatePalette() {
 				g = data[1];
 				b = data[2];
 
-				brightness = (byte)((0.299 * r + 0.587 * g + 0.114 * b) + 0.5);
+				brightness = (byte)((r + g + b) / 3);
 
 				*p++ = brightness;
 				*p++ = brightness;
@@ -1645,7 +1737,7 @@ void ScummEngine::updatePalette() {
 #ifdef USE_RGB_COLOR
 #ifndef DISABLE_TOWNS_DUAL_LAYER_MODE
 	if (_game.platform == Common::kPlatformFMTowns) {
-		p = palette_colors;
+		p = paletteColors;
 		for (int i = first; i < first + num; ++i) {
 			_16BitPalette[i] = get16BitColor(p[0], p[1], p[2]);
 			p += 3;
@@ -1655,7 +1747,36 @@ void ScummEngine::updatePalette() {
 #endif
 #endif
 
-	_system->getPaletteManager()->setPalette(palette_colors, first, num);
+	if (_game.platform == Common::kPlatformMacintosh && _game.heversion == 0 && _useGammaCorrection) {
+		for (int i = 0; i < 3 * num; ++i)
+			paletteColors[i] = Graphics::macGammaCorrectionLookUp[paletteColors[i]];
+	} else if (_game.platform == Common::kPlatformSegaCD && _game.id == GID_MONKEY && _enableSegaShadowMode) {
+		// Apparently the Sega had only 15 levels of intensity for each
+		// color component. You might think there would be 16, but the
+		// palette uses only three bits per color. These can then be
+		// rendered as normal, shadow, or highlight mode, and that comes
+		// out to 15 possible levels.
+
+		// These color levels come from the BlastEm emulator.
+
+		const byte levels[] = { 0, 27, 49, 71, 87, 103, 119, 130, 146, 157, 174, 190, 206, 228, 255 };
+
+		// For reasons unknown, the orignal interpreter rendered
+		// everything in shadow mode. We could easily emulate the other
+		// two modes as well:
+		//
+		// Normal: idx = (paletteColors[i] >> 4) & 0x0E;
+		// Shadow: idx = (paletteColors[i] >> 5) & 0x07;
+		// Hilite: idx = ((paletteColors[i] >> 5) & 0x07) + 7;
+
+		for (int i = 0; i < 3 * num; ++i)
+			paletteColors[i] = levels[(paletteColors[i] >> 5) & 0x07];
+	}
+
+	_system->getPaletteManager()->setPalette(paletteColors, first, num);
+
+	if (_macGui)
+		_macGui->setPaletteDirty();
 }
 
 } // End of namespace Scumm
