@@ -29,11 +29,11 @@
 #define GRAPHICS_TINYGL_ZBUFFER_H_
 
 #include "graphics/surface.h"
-#include "graphics/tinygl/pixelbuffer.h"
 #include "graphics/tinygl/texelbuffer.h"
 #include "graphics/tinygl/gl.h"
 
 #include "common/rect.h"
+#include "common/textconsole.h"
 
 namespace TinyGL {
 
@@ -76,6 +76,8 @@ static const int DRAW_DEPTH_ONLY = 0;
 static const int DRAW_FLAT = 1;
 static const int DRAW_SMOOTH = 2;
 
+struct GLTextureEnv; // defined in zdirtyrect.h
+
 struct Buffer {
 	byte *pbuf;
 	uint *zbuf;
@@ -112,7 +114,7 @@ struct FrameBuffer {
 	}
 
 	byte *getPixelBuffer() {
-		return _pbuf.getRawBuffer();
+		return _pbuf;
 	}
 
 	int getPixelBufferWidth() {
@@ -123,22 +125,71 @@ struct FrameBuffer {
 		return _pbufHeight;
 	}
 
+	int getPixelBufferPitch() {
+		return _pbufPitch;
+	}
+
+	int getPixelBufferBpp() {
+		return _pbufBpp;
+	}
+
 	const uint *getZBuffer() {
 		return _zbuf;
 	}
 
-	Graphics::Surface *copyToBuffer(const Graphics::PixelFormat &dstFormat) {
+	Graphics::Surface *copyFromFrameBuffer(const Graphics::PixelFormat &dstFormat) {
 		Graphics::Surface tmp;
-		tmp.init(_pbufWidth, _pbufHeight, _pbufPitch, _pbuf.getRawBuffer(), _pbufFormat);
+		tmp.init(_pbufWidth, _pbufHeight, _pbufPitch, _pbuf, _pbufFormat);
 		return tmp.convertTo(dstFormat);
 	}
 
 	void getSurfaceRef(Graphics::Surface &surface) {
-		surface.init(_pbufWidth, _pbufHeight, _pbufPitch, _pbuf.getRawBuffer(), _pbufFormat);
+		surface.init(_pbufWidth, _pbufHeight, _pbufPitch, _pbuf, _pbufFormat);
 	}
 
 private:
 
+	FORCEINLINE void setPixelAt(int pixel, uint32 value) {
+		switch (_pbufBpp) {
+		case 2:
+			((uint16 *) _pbuf)[pixel] = value;
+			return;
+		case 3:
+			pixel *= 3;
+#if defined(SCUMM_BIG_ENDIAN)
+			_pbuf[pixel + 0] = (value >> 16) & 0xFF;
+			_pbuf[pixel + 1] = (value >> 8) & 0xFF;
+			_pbuf[pixel + 2] = value & 0xFF;
+#elif defined(SCUMM_LITTLE_ENDIAN)
+			_pbuf[pixel + 0] = value & 0xFF;
+			_pbuf[pixel + 1] = (value >> 8) & 0xFF;
+			_pbuf[pixel + 2] = (value >> 16) & 0xFF;
+#endif
+			return;
+		case 4:
+			((uint32 *) _pbuf)[pixel] = value;
+			return;
+		}
+		error("setPixelAt: Unhandled bytesPerPixel %d", int(_pbufBpp));
+	}
+
+	FORCEINLINE uint32 getPixelAt(int i) const {
+		switch (_pbufBpp) {
+		case 2:
+			return ((uint16 *) _pbuf)[i];
+		case 3:
+			i *= 3;
+#if defined(SCUMM_BIG_ENDIAN)
+			return (_pbuf[i + 0] << 16) | (_pbuf[i + 1] << 8) | _pbuf[i + 2];
+#elif defined(SCUMM_LITTLE_ENDIAN)
+			return _pbuf[i + 0] | (_pbuf[i + 1] << 8) | (_pbuf[i + 2] << 16);
+#endif
+		case 4:
+			return ((uint32 *) _pbuf)[i];
+		}
+		error("getPixelAt: Unhandled bytesPerPixel %d", int(_pbufBpp));
+	}
+
 	FORCEINLINE bool compareDepth(uint &zSrc, uint &zDst) {
 		if (!_depthTestEnabled)
 			return true;
@@ -249,34 +300,35 @@ private:
 
 	FORCEINLINE void stencilOp(bool stencilTestResult, bool depthTestResult, byte *sDst) {
 		int op = !stencilTestResult ? _stencilSfail : !depthTestResult ? _stencilDpfail : _stencilDppass;
-		byte value = *sDst;
+		byte oldValue = *sDst;
+		byte newValue = oldValue;
 		switch (op) {
 		case TGL_KEEP:
 			return;
 		case TGL_ZERO:
-			value = 0;
+			newValue = 0;
 			break;
 		case TGL_REPLACE:
-			value = _stencilRefVal;
+			newValue = _stencilRefVal;
 			break;
 		case TGL_INCR:
-			if (value < 255)
-				value++;
+			if (newValue < 255)
+				newValue++;
 			break;
 		case TGL_INCR_WRAP:
-			value++;
+			newValue++;
 			break;
 		case TGL_DECR:
-			if (value > 0)
-				value--;
+			if (newValue > 0)
+				newValue--;
 			break;
 		case TGL_DECR_WRAP:
-			value--;
+			newValue--;
 			break;
 		case TGL_INVERT:
-			value = ~value;
+			newValue = ~newValue;
 		}
-		*sDst = value & _stencilWriteMask;
+		*sDst = (newValue & _stencilWriteMask) | (oldValue & ~_stencilWriteMask);
 	}
 
 	template <bool kEnableAlphaTest, bool kBlendingEnabled>
@@ -287,13 +339,13 @@ private:
 	template <bool kEnableAlphaTest, bool kBlendingEnabled, bool kDepthWrite>
 	FORCEINLINE void writePixel(int pixel, int value, uint z) {
 		if (kBlendingEnabled == false) {
-			_pbuf.setPixelAt(pixel, value);
+			setPixelAt(pixel, value);
 			if (kDepthWrite) {
 				_zbuf[pixel] = z;
 			}
 		} else {
 			byte rSrc, gSrc, bSrc, aSrc;
-			_pbuf.getFormat().colorToARGB(value, aSrc, rSrc, gSrc, bSrc);
+			_pbufFormat.colorToARGB(value, aSrc, rSrc, gSrc, bSrc);
 
 			writePixel<kEnableAlphaTest, kBlendingEnabled, kDepthWrite>(pixel, aSrc, rSrc, gSrc, bSrc, z);
 		}
@@ -307,13 +359,19 @@ private:
 		}
 	}
 
-	template <bool kDepthWrite, bool kSmoothMode, bool kFogMode, bool kEnableAlphaTest, bool kEnableScissor, bool kEnableBlending, bool kStencilEnabled, bool kDepthTestEnabled>
+	template <bool kDepthWrite, bool kSmoothMode, bool kFogMode, bool kEnableAlphaTest, bool kEnableScissor, bool kEnableBlending, bool kStencilEnabled, bool kStippleEnabled, bool kDepthTestEnabled>
 	void putPixelNoTexture(int fbOffset, uint *pz, byte *ps, int _a,
 	                       int x, int y, uint &z, uint &r, uint &g, uint &b, uint &a,
 	                       int &dzdx, int &drdx, int &dgdx, int &dbdx, uint dadx,
 	                       uint &fog, int fog_r, int fog_g, int fog_b, int &dfdx);
 
-	template <bool kDepthWrite, bool kLightsMode, bool kSmoothMode, bool kFogMode, bool kEnableAlphaTest, bool kEnableScissor, bool kEnableBlending, bool kStencilEnabled, bool kDepthTestEnabled>
+	enum class ColorMode {
+		NoInterpolation,
+		Default, // GL_TEXTURE_ENV_MODE == GL_MODULATE
+		CustomTexEnv
+	};
+
+	template <bool kDepthWrite, ColorMode kColorMode, bool kSmoothMode, bool kFogMode, bool kEnableAlphaTest, bool kEnableScissor, bool kEnableBlending, bool kStencilEnabled, bool kDepthTestEnabled>
 	void putPixelTexture(int fbOffset, const TexelBuffer *texture,
 	                     uint wrap_s, uint wrap_t, uint *pz, byte *ps, int _a,
 	                     int x, int y, uint &z, int &t, int &s,
@@ -321,7 +379,7 @@ private:
 	                     int &dzdx, int &dsdx, int &dtdx, int &drdx, int &dgdx, int &dbdx, uint dadx,
 	                     uint &fog, int fog_r, int fog_g, int fog_b, int &dfdx);
 
-	template <bool kDepthWrite, bool kEnableScissor, bool kStencilEnabled, bool kDepthTestEnabled>
+	template <bool kDepthWrite, bool kEnableScissor, bool kStencilEnabled, bool kStippleEnabled, bool kDepthTestEnabled>
 	void putPixelDepth(uint *pz, byte *ps, int _a, int x, int y, uint &z, int &dzdx);
 
 
@@ -370,7 +428,7 @@ private:
 
 	template <bool kEnableAlphaTest, bool kBlendingEnabled, bool kDepthWrite>
 	FORCEINLINE void writePixel(int pixel, byte aSrc, byte rSrc, byte gSrc, byte bSrc, uint z) {
-		writePixel<kEnableAlphaTest, kBlendingEnabled, false, false>(pixel, aSrc, rSrc, gSrc, bSrc, z, 0.0f, 0, 0, 0);
+		writePixel<kEnableAlphaTest, kBlendingEnabled, kDepthWrite, false>(pixel, aSrc, rSrc, gSrc, bSrc, z, 0.0f, 0, 0, 0);
 	}
 
 	template <bool kEnableAlphaTest, bool kBlendingEnabled, bool kDepthWrite, bool kFogMode>
@@ -407,10 +465,10 @@ private:
 		}
 
 		if (!kBlendingEnabled) {
-			_pbuf.setPixelAt(pixel, aSrc, rSrc, gSrc, bSrc);
+			setPixelAt(pixel, _pbufFormat.ARGBToColor(aSrc, rSrc, gSrc, bSrc));
 		} else {
 			byte rDst, gDst, bDst, aDst;
-			_pbuf.getARGBAt(pixel, aDst, rDst, gDst, bDst);
+			_pbufFormat.colorToARGB(getPixelAt(pixel), aDst, rDst, gDst, bDst);
 			switch (_sourceBlendingFactor) {
 			case TGL_ZERO:
 				rSrc = gSrc = bSrc = 0;
@@ -509,7 +567,7 @@ private:
 			if (finalB > 255) {
 				finalB = 255;
 			}
-			_pbuf.setPixelAt(pixel, 255, finalR, finalG, finalB);
+			setPixelAt(pixel, _pbufFormat.RGBToColor(finalR, finalG, finalB));
 		}
 	}
 
@@ -520,13 +578,32 @@ public:
 	void clearRegion(int x, int y, int w, int h, bool clearZ, int z,
 	                 bool clearColor, int r, int g, int b, bool clearStencil, int stencilValue);
 
-	void setScissorRectangle(const Common::Rect &rect) {
-		_clipRectangle = rect;
-		_enableScissor = true;
+	const Common::Rect &getClippingRectangle() const {
+		return _clipRectangle;
 	}
 
-	void resetScissorRectangle() {
-		_enableScissor = false;
+	void setupScissor(bool enable, const int (&scissor)[4], const Common::Rect *clippingRectangle) {
+		_clippingEnabled = enable || clippingRectangle;
+
+		if (enable && clippingRectangle) {
+			_clipRectangle = clippingRectangle->findIntersectingRect(Common::Rect(
+					scissor[0],
+					// all viewport calculations are already flipped upside down
+					_pbufHeight - scissor[1] - scissor[3],
+					scissor[0] + scissor[2],
+					_pbufHeight - scissor[1]));
+		} else if (enable) {
+			_clipRectangle = Common::Rect(
+					scissor[0],
+					// all viewport calculations are already flipped upside down
+					_pbufHeight - scissor[1] - scissor[3],
+					scissor[0] + scissor[2],
+					_pbufHeight - scissor[1]);
+		} else if (clippingRectangle) {
+			_clipRectangle = *clippingRectangle;
+		} else {
+			_clipRectangle = Common::Rect(0, 0, _pbufWidth, _pbufHeight);
+		}
 	}
 
 	void enableBlending(bool enable) {
@@ -567,7 +644,23 @@ public:
 		_stencilWriteMask = stencilWriteMask;
 	}
 
-	void setStencilTestFunc(int stencilFunc, int stencilValue, uint stencilMask) {
+	void enablePolygonStipple(bool enable) {
+		_polygonStippleEnabled = enable;
+	}
+
+	void setPolygonStipplePattern(const byte *stipple) {
+		_polygonStipplePattern = stipple;
+	}
+
+	void enableTwoColorStipple(bool enable) {
+		_twoColorStippleEnabled = enable;
+	}
+
+	void setStippleColor(int r, int g, int b) {
+		_stippleColor = RGB_TO_PIXEL(r, g, b);
+	}
+
+	void setStencilTestFunc(int stencilFunc, byte stencilValue, byte stencilMask) {
 		_stencilTestFunc = stencilFunc;
 		_stencilRefVal = stencilValue;
 		_stencilMask = stencilMask;
@@ -597,6 +690,10 @@ public:
 		_wrapT = wrapt;
 	}
 
+	void setTextureEnvironment(const GLTextureEnv *texEnv) {
+		_textureEnv = texEnv;
+	}
+
 	void setTextureSizeAndMask(int textureSize, int textureSizeMask) {
 		_textureSize = textureSize;
 		_textureSizeMask = textureSizeMask;
@@ -625,39 +722,47 @@ private:
 	void selectOffscreenBuffer(Buffer *buffer);
 	void clearOffscreenBuffer(Buffer *buffer);
 
-	template <bool kInterpRGB, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
+	template <ColorMode kColorMode, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
 	          bool kDepthWrite, bool kFogMode, bool kAlphaTestEnabled, bool kEnableScissor,
-	          bool kBlendingEnabled, bool kStencilEnabled, bool kDepthTestEnabled>
+	          bool kBlendingEnabled, bool kStencilEnabled, bool kStippleEnabled, bool kDepthTestEnabled>
 	void fillTriangle(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
 
-	template <bool kInterpRGB, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
+	template <ColorMode kColorMode, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
+	          bool kDepthWrite, bool kFogMode, bool kAlphaTestEnabled, bool kEnableScissor,
+	          bool kBlendingEnabled, bool kStencilEnabled, bool kStippleEnabled>
+	void fillTriangle(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
+
+	template <ColorMode kColorMode, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
 	          bool kDepthWrite, bool kFogMode, bool kAlphaTestEnabled, bool kEnableScissor,
 	          bool kBlendingEnabled, bool kStencilEnabled>
 	void fillTriangle(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
 
-	template <bool kInterpRGB, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
+	template <ColorMode kColorMode, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
 	          bool kDepthWrite, bool kFogMode, bool enableAlphaTest, bool kEnableScissor, bool kBlendingEnabled>
 	void fillTriangle(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
 
-	template <bool kInterpRGB, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
+	template <ColorMode kColorMode, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
 	          bool kDepthWrite, bool kFogMode, bool enableAlphaTest, bool kEnableScissor>
 	void fillTriangle(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
 
-	template <bool kInterpRGB, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
+	template <ColorMode kColorMode, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
 	          bool kDepthWrite, bool kFogMode, bool enableAlphaTest>
 	void fillTriangle(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
 
-	template <bool kInterpRGB, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
+	template <ColorMode kColorMode, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
 	          bool kDepthWrite, bool kFogMode>
 	void fillTriangle(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
 
-	template <bool kInterpRGB, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
+	template <ColorMode kColorMode, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode,
 	          bool kDepthWrite>
 	void fillTriangle(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
 
-	template <bool kInterpRGB, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode>
+	template <ColorMode kColorMode, bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode>
 	void fillTriangle(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
 
+	template <bool kInterpZ, bool kInterpST, bool kInterpSTZ, bool kSmoothMode, bool kDepthWrite>
+	void fillTriangleTextureMapping(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
+
 public:
 
 	void fillTriangleTextureMappingPerspectiveSmooth(ZBufferPoint *p0, ZBufferPoint *p1, ZBufferPoint *p2);
@@ -692,8 +797,18 @@ private:
 	template <bool kInterpRGB, bool kInterpZ, bool kDepthWrite, bool kEnableScissor>
 	void drawLine(const ZBufferPoint *p1, const ZBufferPoint *p2);
 
+	void applyTextureEnvironment(
+		int internalformat,
+		uint previousA, uint previousR, uint previousG, uint previousB,
+		byte &texA, byte &texR, byte &texG, byte &texB);
+
+	// the same as GL_TEXTURE_ENV_MODE == GL_MODULATE as fast-path for the default mode
+	void applyModulation(
+		uint previousA, uint previousR, uint previousG, uint previousB,
+		byte &texA, byte &texR, byte &texG, byte &texB);
+
 	Buffer _offscreenBuffer;
-	Graphics::PixelBuffer _pbuf;
+	byte *_pbuf;
 	int _pbufWidth;
 	int _pbufHeight;
 	int _pbufPitch;
@@ -708,9 +823,10 @@ private:
 	int _textureSizeMask;
 
 	Common::Rect _clipRectangle;
-	bool _enableScissor;
+	bool _clippingEnabled;
 
 	const TexelBuffer *_currentTexture;
+	const GLTextureEnv *_textureEnv;
 	uint _wrapS, _wrapT;
 	bool _blendingEnabled;
 	int _sourceBlendingFactor;
@@ -722,12 +838,17 @@ private:
 	bool _depthWrite;
 	bool _stencilTestEnabled;
 	int _stencilTestFunc;
-	int _stencilRefVal;
-	uint _stencilMask;
-	uint _stencilWriteMask;
+	byte _stencilRefVal;
+	byte _stencilMask;
+	byte _stencilWriteMask;
 	int _stencilSfail;
 	int _stencilDpfail;
 	int _stencilDppass;
+
+	bool _polygonStippleEnabled;
+	bool _twoColorStippleEnabled;
+	uint32 _stippleColor;
+	const byte *_polygonStipplePattern;
 	int _depthFunc;
 	int _offsetStates;
 	float _offsetFactor;
@@ -742,6 +863,7 @@ private:
 void gl_free(void *p);
 void *gl_malloc(int size);
 void *gl_zalloc(int size);
+void *gl_realloc(void *p, int size);
 
 } // end of namespace TinyGL
 
