@@ -4,10 +4,10 @@
  * are too numerous to list here. Please refer to the COPYRIGHT
  * file distributed with this source distribution.
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * This program is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -15,8 +15,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  */
 
@@ -33,16 +32,60 @@
 
 namespace GUI {
 
-ListWidget::ListWidget(Dialog *boss, const String &name, const char *tooltip, uint32 cmd)
+bool ListWidgetDefaultMatcher(void *, int, const Common::U32String &item, const Common::U32String &token) {
+	return item.contains(token);
+}
+
+ListWidget::ListWidget(Dialog *boss, const Common::String &name, const Common::U32String &tooltip, uint32 cmd)
 	: EditableWidget(boss, name, tooltip), _cmd(cmd) {
 
-	_scrollBar = NULL;
-	_textWidth = NULL;
+	_entriesPerPage = 0;
+	_scrollBarWidth = 0;
 
-	// This ensures that _entriesPerPage is properly initialized.
-	reflowLayout();
+	_scrollBar = new ScrollBarWidget(this, _w - _scrollBarWidth, 0, _scrollBarWidth, _h);
+	_scrollBar->setTarget(this);
 
-	_scrollBar = new ScrollBarWidget(this, _w - _scrollBarWidth + 1, 0, _scrollBarWidth, _h);
+	setFlags(WIDGET_ENABLED | WIDGET_CLEARBG | WIDGET_RETAIN_FOCUS | WIDGET_WANT_TICKLE | WIDGET_TRACK_MOUSE);
+	_type = kListWidget;
+	_editMode = false;
+	_numberingMode = kListNumberingOne;
+	_currentPos = 0;
+	_selectedItem = -1;
+	_currentKeyDown = 0;
+
+	_quickSelectTime = 0;
+
+	// The item is selected, thus _bgcolor is used to draw the caret and _textcolorhi to erase it
+	_caretInverse = true;
+
+	// Disable text selection
+	_disableSelection = true;
+
+	// FIXME: This flag should come from widget definition
+	_editable = true;
+
+	_quickSelect = true;
+	_editColor = ThemeEngine::kFontColorNormal;
+	_dictionarySelect = false;
+
+	_filterMatcher = ListWidgetDefaultMatcher;
+	_filterMatcherArg = nullptr;
+
+	_lastRead = -1;
+
+	_hlLeftPadding = _hlRightPadding = 0;
+	_leftPadding = _rightPadding = 0;
+	_topPadding = _bottomPadding = 0;
+	_itemSpacing = 0;
+}
+
+ListWidget::ListWidget(Dialog *boss, int x, int y, int w, int h, bool scale, const Common::U32String &tooltip, uint32 cmd)
+	: EditableWidget(boss, x, y, w, h, scale, tooltip), _cmd(cmd) {
+
+	_entriesPerPage = 0;
+	_scrollBarWidth = 0;
+
+	_scrollBar = new ScrollBarWidget(this, _w - _scrollBarWidth, 0, _scrollBarWidth, _h);
 	_scrollBar->setTarget(this);
 
 	setFlags(WIDGET_ENABLED | WIDGET_CLEARBG | WIDGET_RETAIN_FOCUS | WIDGET_WANT_TICKLE);
@@ -58,47 +101,52 @@ ListWidget::ListWidget(Dialog *boss, const String &name, const char *tooltip, ui
 	// The item is selected, thus _bgcolor is used to draw the caret and _textcolorhi to erase it
 	_caretInverse = true;
 
-	// FIXME: This flag should come from widget definition
-	_editable = true;
-
-	_quickSelect = true;
-	_editColor = ThemeEngine::kFontColorNormal;
-}
-
-ListWidget::ListWidget(Dialog *boss, int x, int y, int w, int h, const char *tooltip, uint32 cmd)
-	: EditableWidget(boss, x, y, w, h, tooltip), _cmd(cmd) {
-
-	_scrollBar = NULL;
-	_textWidth = NULL;
-
-	// This ensures that _entriesPerPage is properly initialized.
-	reflowLayout();
-
-	_scrollBar = new ScrollBarWidget(this, _w - _scrollBarWidth + 1, 0, _scrollBarWidth, _h);
-	_scrollBar->setTarget(this);
-
-	setFlags(WIDGET_ENABLED | WIDGET_CLEARBG | WIDGET_RETAIN_FOCUS | WIDGET_WANT_TICKLE);
-	_type = kListWidget;
-	_editMode = false;
-	_numberingMode = kListNumberingOne;
-	_currentPos = 0;
-	_selectedItem = -1;
-	_currentKeyDown = 0;
-
-	_quickSelectTime = 0;
-
-	// The item is selected, thus _bgcolor is used to draw the caret and _textcolorhi to erase it
-	_caretInverse = true;
+	// Disable text selection
+	_disableSelection = true;
 
 	// FIXME: This flag should come from widget definition
 	_editable = true;
 
 	_quickSelect = true;
 	_editColor = ThemeEngine::kFontColorNormal;
+	_dictionarySelect = false;
+
+	_filterMatcher = ListWidgetDefaultMatcher;
+	_filterMatcherArg = nullptr;
+
+	_lastRead = -1;
+
+	_hlLeftPadding = _hlRightPadding = 0;
+	_leftPadding = _rightPadding = 0;
+	_topPadding = _bottomPadding = 0;
+	_itemSpacing = 0;
+
+	_scrollBarWidth = 0;
 }
 
-ListWidget::~ListWidget() {
-	delete[] _textWidth;
+ListWidget::ListWidget(Dialog *boss, int x, int y, int w, int h, const Common::U32String &tooltip, uint32 cmd)
+	: ListWidget(boss, x, y, w, h, false, tooltip, cmd) {
+}
+
+void ListWidget::copyListData(const Common::U32StringArray &list) {
+	Common::U32String stripped;
+
+	_dataList.clear();
+	_cleanedList.clear();
+
+	for (uint i = 0; i < list.size(); ++i) {
+		stripped = stripGUIformatting(list[i]);
+
+		_dataList.push_back(ListData(list[i], stripped));
+		_cleanedList.push_back(stripped);
+	}
+}
+
+
+bool ListWidget::containsWidget(Widget *w) const {
+	if (w == _scrollBar || _scrollBar->containsWidget(w))
+		return true;
+	return false;
 }
 
 Widget *ListWidget::findWidget(int x, int y) {
@@ -112,7 +160,7 @@ void ListWidget::setSelected(int item) {
 	// HACK/FIXME: If our _listIndex has a non zero size,
 	// we will need to look up, whether the user selected
 	// item is present in that list
-	if (_listIndex.size()) {
+	if (!_filter.empty()) {
 		int filteredItem = -1;
 
 		for (uint i = 0; i < _listIndex.size(); ++i) {
@@ -137,37 +185,24 @@ void ListWidget::setSelected(int item) {
 		// Notify clients that the selection changed.
 		sendCommand(kListSelectionChangedCmd, _selectedItem);
 
-		_currentPos = _selectedItem - _entriesPerPage / 2;
-		scrollToCurrent();
-		draw();
+		if (!isItemVisible(_selectedItem)) {
+			// scroll selected item to center if possible
+			_currentPos = _selectedItem - _entriesPerPage / 2;
+			scrollToCurrent();
+		}
+		markAsDirty();
 	}
 }
 
-ThemeEngine::FontColor ListWidget::getSelectionColor() const {
-	if (_listColors.empty())
-		return ThemeEngine::kFontColorNormal;
-
-	if (_filter.empty())
-		return _listColors[_selectedItem];
-	else
-		return _listColors[_listIndex[_selectedItem]];
-}
-
-void ListWidget::setList(const StringArray &list, const ColorList *colors) {
+void ListWidget::setList(const Common::U32StringArray &list) {
 	if (_editMode && _caretVisible)
 		drawCaret(true);
 
 	// Copy everything
-	_dataList = list;
+	copyListData(list);
 	_list = list;
 	_filter.clear();
 	_listIndex.clear();
-	_listColors.clear();
-
-	if (colors) {
-		_listColors = *colors;
-		assert(_listColors.size() == _dataList.size());
-	}
 
 	int size = list.size();
 	if (_currentPos >= size)
@@ -180,20 +215,10 @@ void ListWidget::setList(const StringArray &list, const ColorList *colors) {
 	scrollBarRecalc();
 }
 
-void ListWidget::append(const String &s, ThemeEngine::FontColor color) {
-	if (_dataList.size() == _listColors.size()) {
-		// If the color list has the size of the data list, we append the color.
-		_listColors.push_back(color);
-	} else if (!_listColors.size() && color != ThemeEngine::kFontColorNormal) {
-		// If it's the first entry to use a non default color, we will fill
-		// up all other entries of the color list with the default color and
-		// add the requested color for the new entry.
-		for (uint i = 0; i < _dataList.size(); ++i)
-			_listColors.push_back(ThemeEngine::kFontColorNormal);
-		_listColors.push_back(color);
-	}
-
-	_dataList.push_back(s);
+void ListWidget::append(const Common::String &s) {
+	Common::U32String stripped = stripGUIformatting(s);
+	_dataList.push_back(ListData(s, stripped));
+	_cleanedList.push_back(stripped);
 	_list.push_back(s);
 
 	setFilter(_filter, false);
@@ -242,11 +267,15 @@ void ListWidget::handleMouseDown(int x, int y, int button, int clickCount) {
 		sendCommand(kListSelectionChangedCmd, _selectedItem);
 	}
 
+	// Notify clients if an item was clicked
+	if (newSelectedItem >= 0) {
+		sendCommand(kListItemSingleClickedCmd, _selectedItem);
+	}
+
 	// TODO: Determine where inside the string the user clicked and place the
 	// caret accordingly.
 	// See _editScrollOffset and EditTextWidget::handleMouseDown.
-	draw();
-
+	markAsDirty();
 }
 
 void ListWidget::handleMouseUp(int x, int y, int button, int clickCount) {
@@ -262,19 +291,47 @@ void ListWidget::handleMouseWheel(int x, int y, int direction) {
 	_scrollBar->handleMouseWheel(x, y, direction);
 }
 
+void ListWidget::handleMouseMoved(int x, int y, int button) {
+	if (!isEnabled())
+		return;
+
+	// Determine if we are inside the widget
+	if (x < 0 || x > _w)
+		return;
+
+	// First check whether the selection changed
+	int item = findItem(x, y);
+
+	if (item != -1) {
+		if(_lastRead != item) {
+			read(stripGUIformatting(_list[item]));
+			_lastRead = item;
+		}
+	}
+	else
+		_lastRead = -1;
+}
+
+void ListWidget::handleMouseLeft(int button) {
+	_lastRead = -1;
+}
+
 
 int ListWidget::findItem(int x, int y) const {
 	if (y < _topPadding) return -1;
-	int item = (y - _topPadding) / kLineHeight + _currentPos;
-	if (item >= _currentPos && item < _currentPos + _entriesPerPage &&
-		item < (int)_list.size())
+	int item = (y - _topPadding) / (kLineHeight + _itemSpacing) + _currentPos;
+	if (isItemVisible(item) && item < (int)_list.size())
 		return item;
 	else
 		return -1;
 }
 
-static int matchingCharsIgnoringCase(const char *x, const char *y, bool &stop) {
+static int matchingCharsIgnoringCase(const char *x, const char *y, bool &stop, bool dictionary) {
 	int match = 0;
+	if (dictionary) {
+		x = scumm_skipArticle(x);
+		y = scumm_skipArticle(y);
+	}
 	while (*x && *y && tolower(*x) == tolower(*y)) {
 		++x;
 		++y;
@@ -309,8 +366,8 @@ bool ListWidget::handleKeyDown(Common::KeyState state) {
 			int newSelectedItem = 0;
 			int bestMatch = 0;
 			bool stop;
-			for (StringArray::const_iterator i = _list.begin(); i != _list.end(); ++i) {
-				const int match = matchingCharsIgnoringCase(i->c_str(), _quickSelectStr.c_str(), stop);
+			for (const auto &entry : _list) {
+				const int match = matchingCharsIgnoringCase(stripGUIformatting(entry).encode().c_str(), _quickSelectStr.c_str(), stop, _dictionarySelect);
 				if (match > bestMatch || stop) {
 					_selectedItem = newSelectedItem;
 					bestMatch = match;
@@ -353,14 +410,11 @@ bool ListWidget::handleKeyDown(Common::KeyState state) {
 				handled = false;
 				break;
 			}
+			// fall through
 		case Common::KEYCODE_BACKSPACE:
 		case Common::KEYCODE_DELETE:
 			if (_selectedItem >= 0) {
-				if (_editable) {
-					// Ignore delete and backspace when the list item is editable
-				} else {
-					sendCommand(kListItemRemovalRequestCmd, _selectedItem);
-				}
+				sendCommand(kListItemRemovalRequestCmd, _selectedItem);
 			}
 			break;
 
@@ -369,6 +423,7 @@ bool ListWidget::handleKeyDown(Common::KeyState state) {
 				handled = false;
 				break;
 			}
+			// fall through
 		case Common::KEYCODE_END:
 			_selectedItem = _list.size() - 1;
 			break;
@@ -379,6 +434,7 @@ bool ListWidget::handleKeyDown(Common::KeyState state) {
 				handled = false;
 				break;
 			}
+			// fall through
 		case Common::KEYCODE_DOWN:
 			if (_selectedItem < (int)_list.size() - 1)
 				_selectedItem++;
@@ -389,6 +445,7 @@ bool ListWidget::handleKeyDown(Common::KeyState state) {
 				handled = false;
 				break;
 			}
+			// fall through
 		case Common::KEYCODE_PAGEDOWN:
 			_selectedItem += _entriesPerPage - 1;
 			if (_selectedItem >= (int)_list.size() )
@@ -400,6 +457,7 @@ bool ListWidget::handleKeyDown(Common::KeyState state) {
 				handled = false;
 				break;
 			}
+			// fall through
 		case Common::KEYCODE_HOME:
 			_selectedItem = 0;
 			break;
@@ -409,6 +467,7 @@ bool ListWidget::handleKeyDown(Common::KeyState state) {
 				handled = false;
 				break;
 			}
+			// fall through
 		case Common::KEYCODE_UP:
 			if (_selectedItem > 0)
 				_selectedItem--;
@@ -419,6 +478,7 @@ bool ListWidget::handleKeyDown(Common::KeyState state) {
 				handled = false;
 				break;
 			}
+			// fall through
 		case Common::KEYCODE_PAGEUP:
 			_selectedItem -= _entriesPerPage - 1;
 			if (_selectedItem < 0)
@@ -433,12 +493,12 @@ bool ListWidget::handleKeyDown(Common::KeyState state) {
 	}
 
 	if (dirty || _selectedItem != oldSelectedItem)
-		draw();
+		markAsDirty();
 
 	if (_selectedItem != oldSelectedItem) {
 		sendCommand(kListSelectionChangedCmd, _selectedItem);
 		// also draw scrollbar
-		_scrollBar->draw();
+		_scrollBar->markAsDirty();
 	}
 
 	return handled;
@@ -454,7 +514,7 @@ void ListWidget::receivedFocusWidget() {
 	_inversion = ThemeEngine::kTextInversionFocus;
 
 	// Redraw the widget so the selection color will change
-	draw();
+	markAsDirty();
 }
 
 void ListWidget::lostFocusWidget() {
@@ -464,7 +524,7 @@ void ListWidget::lostFocusWidget() {
 	_editMode = false;
 	g_system->setFeatureState(OSystem::kFeatureVirtualKeyboard, false);
 	drawCaret(true);
-	draw();
+	markAsDirty();
 }
 
 void ListWidget::handleCommand(CommandSender *sender, uint32 cmd, uint32 data) {
@@ -473,77 +533,107 @@ void ListWidget::handleCommand(CommandSender *sender, uint32 cmd, uint32 data) {
 		if (_currentPos != (int)data) {
 			_currentPos = data;
 			checkBounds();
-			draw();
+			markAsDirty();
 
 			// Scrollbar actions cause list focus (which triggers a redraw)
 			// NOTE: ListWidget's boss is always GUI::Dialog
 			((GUI::Dialog *)_boss)->setFocusWidget(this);
 		}
 		break;
+	default:
+		break;
 	}
 }
 
 void ListWidget::drawWidget() {
 	int i, pos, len = _list.size();
-	Common::String buffer;
+	Common::U32String buffer;
 
 	// Draw a thin frame around the list.
-	g_gui.theme()->drawWidgetBackground(Common::Rect(_x, _y, _x + _w, _y + _h), 0, ThemeEngine::kWidgetBackgroundBorder);
-	const int scrollbarW = (_scrollBar && _scrollBar->isVisible()) ? _scrollBarWidth : 0;
+	g_gui.theme()->drawWidgetBackground(Common::Rect(_x, _y, _x + _w, _y + _h),
+	                                    ThemeEngine::kWidgetBackgroundBorder);
 
 	// Draw the list items
+	const int lineHeight = kLineHeight + _itemSpacing;
 	for (i = 0, pos = _currentPos; i < _entriesPerPage && pos < len; i++, pos++) {
-		const int y = _y + _topPadding + kLineHeight * i;
-		const int fontHeight = kLineHeight;
+		const int y = _y + _topPadding + lineHeight * i;
 		ThemeEngine::TextInversionState inverted = ThemeEngine::kTextInversionNone;
 
 		// Draw the selected item inverted, on a highlighted background.
 		if (_selectedItem == pos)
 			inverted = _inversion;
 
+		// Get state for drawing the item text
+		ThemeEngine::WidgetStateInfo itemState = getItemState(pos);
+
 		Common::Rect r(getEditRect());
 		int pad = _leftPadding;
+		int rtlPad = (_x + r.left + _leftPadding) - (_x + _hlLeftPadding);
 
-		// If in numbering mode, we first print a number prefix
-		if (_numberingMode != kListNumberingOff) {
+		// If in numbering mode & not in RTL based GUI, we first print a number prefix
+		if (_numberingMode != kListNumberingOff && g_gui.useRTL() == false) {
 			buffer = Common::String::format("%2d. ", (pos + _numberingMode));
-			g_gui.theme()->drawText(Common::Rect(_x, y, _x + r.left + _leftPadding, y + fontHeight - 2),
-									buffer, _state, Graphics::kTextAlignLeft, inverted, _leftPadding, true);
+			g_gui.theme()->drawText(Common::Rect(_x + _hlLeftPadding, y, _x + r.left + _leftPadding, y + lineHeight),
+									buffer, itemState, _drawAlign, inverted, _leftPadding, true);
 			pad = 0;
 		}
 
-		int width;
+		Common::Rect r1(_x + r.left, y, _x + r.right, y + lineHeight);
 
-		ThemeEngine::FontColor color = ThemeEngine::kFontColorNormal;
+		if (g_gui.useRTL()) {
+			if (_scrollBar->isVisible()) {
+				r1.translate(_scrollBarWidth, 0);
+			}
 
-		if (!_listColors.empty()) {
-			if (_filter.empty() || _selectedItem == -1)
-				color = _listColors[pos];
-			else
-				color = _listColors[_listIndex[pos]];
+			if (_numberingMode != kListNumberingOff) {
+				r1.translate(-rtlPad, 0);
+			}
 		}
 
+		ThemeEngine::FontColor color = ThemeEngine::kFontColorFormatting;
+
 		if (_selectedItem == pos && _editMode) {
 			buffer = _editString;
 			color = _editColor;
 			adjustOffset();
-			width = _w - r.left - _hlRightPadding - _leftPadding - scrollbarW;
-			g_gui.theme()->drawText(Common::Rect(_x + r.left, y, _x + r.left + width, y + fontHeight - 2), buffer, _state,
-									Graphics::kTextAlignLeft, inverted, pad, true, ThemeEngine::kFontStyleBold, color);
 		} else {
 			buffer = _list[pos];
-			width = _w - r.left - scrollbarW;
-			g_gui.theme()->drawText(Common::Rect(_x + r.left, y, _x + r.left + width, y + fontHeight - 2), buffer, _state,
-									Graphics::kTextAlignLeft, inverted, pad, true, ThemeEngine::kFontStyleBold, color);
 		}
 
-		_textWidth[i] = width;
+		drawFormattedText(r1, buffer, itemState, _drawAlign, inverted, pad, true, color);
+
+		// If in numbering mode & using RTL layout in GUI, we print a number suffix after drawing the text
+		if (_numberingMode != kListNumberingOff && g_gui.useRTL()) {
+			buffer = Common::String::format(" .%2d", (pos + _numberingMode));
+
+			Common::Rect r2 = r1;
+
+			r2.left = r1.right;
+			r2.right = r1.right + rtlPad;
+
+			g_gui.theme()->drawText(r2, buffer, itemState, _drawAlign, inverted, _leftPadding, true);
+		}
+	}
+
+	if (_editMode) {
+		EditableWidget::drawWidget();
 	}
 }
 
 Common::Rect ListWidget::getEditRect() const {
-	Common::Rect r(_hlLeftPadding, 0, _w - _hlLeftPadding - _hlRightPadding, kLineHeight - 2);
-	const int offset = (_selectedItem - _currentPos) * kLineHeight + _topPadding;
+	const int scrollbarW = (_scrollBar && _scrollBar->isVisible()) ? _scrollBarWidth : 0;
+	int editWidth = _w - _hlLeftPadding - _hlRightPadding - scrollbarW;
+	// Ensure r will always be a valid rect
+	if (editWidth < 0) {
+		editWidth = 0;
+	}
+
+	const int fontHeight = g_gui.getFontHeight();
+	const int lineHeight = kLineHeight + _itemSpacing;
+
+	// Center the caret/text vertically inside the line
+	Common::Rect r(_hlLeftPadding, 0, _hlLeftPadding + editWidth, fontHeight);
+	const int offset = (_selectedItem - _currentPos) * lineHeight + _topPadding + (lineHeight - fontHeight) / 2;
 	r.top += offset;
 	r.bottom += offset;
 
@@ -551,11 +641,21 @@ Common::Rect ListWidget::getEditRect() const {
 		// FIXME: Assumes that all digits have the same width.
 		Common::String temp = Common::String::format("%2d. ", (_list.size() - 1 + _numberingMode));
 		r.left += g_gui.getStringWidth(temp) + _leftPadding;
+		// Make sure we don't go farther than right
+		if (r.right < r.left) {
+			r.right = r.left;
+		}
 	}
 
 	return r;
 }
 
+int ListWidget::getCaretOffset() const {
+	Common::U32String substr(_editString.begin(), _editString.begin() + _caretPos);
+	Common::U32String stripped = stripGUIformatting(substr);
+	return g_gui.getStringWidth(stripped, _font) - _editScrollOffset;
+}
+
 void ListWidget::checkBounds() {
 	if (_currentPos < 0 || _entriesPerPage > (int)_list.size())
 		_currentPos = 0;
@@ -565,7 +665,7 @@ void ListWidget::checkBounds() {
 
 void ListWidget::scrollToCurrent() {
 	// Only do something if the current item is not in our view port
-	if (_selectedItem < _currentPos) {
+	if (_selectedItem != -1 && _selectedItem < _currentPos) {
 		// it's above our view
 		_currentPos = _selectedItem;
 	} else if (_selectedItem >= _currentPos + _entriesPerPage ) {
@@ -587,24 +687,18 @@ void ListWidget::scrollToEnd() {
 
 	_scrollBar->_currentPos = _currentPos;
 	_scrollBar->recalc();
-	_scrollBar->draw();
+	_scrollBar->markAsDirty();
 }
 
 void ListWidget::startEditMode() {
 	if (_editable && !_editMode && _selectedItem >= 0) {
 		_editMode = true;
-		setEditString(_list[_selectedItem]);
+		setEditString(stripGUIformatting(_list[_selectedItem]));
 		_caretPos = _editString.size();	// Force caret to the *end* of the selection.
-		if (_listColors.empty()) {
-			_editColor = ThemeEngine::kFontColorNormal;
-		} else {
-			if (_filter.empty())
-				_editColor = _listColors[_selectedItem];
-			else
-				_editColor = _listColors[_listIndex[_selectedItem]];
-		}
-		draw();
+		_editColor = ThemeEngine::kFontColorNormal;
+		markAsDirty();
 		g_system->setFeatureState(OSystem::kFeatureVirtualKeyboard, true);
+		sendCommand(kListItemEditModeStartedCmd, _selectedItem);
 	}
 }
 
@@ -622,8 +716,6 @@ void ListWidget::abortEditMode() {
 	// undo any changes made
 	assert(_selectedItem >= 0);
 	_editMode = false;
-	//drawCaret(true);
-	//draw();
 	g_system->setFeatureState(OSystem::kFeatureVirtualKeyboard, false);
 }
 
@@ -636,6 +728,7 @@ void ListWidget::reflowLayout() {
 	_bottomPadding = g_gui.xmlEval()->getVar("Globals.ListWidget.Padding.Bottom", 0);
 	_hlLeftPadding = g_gui.xmlEval()->getVar("Globals.ListWidget.hlLeftPadding", 0);
 	_hlRightPadding = g_gui.xmlEval()->getVar("Globals.ListWidget.hlRightPadding", 0);
+	_itemSpacing = g_gui.xmlEval()->getVar("Globals.ListWidget.itemSpacing", 0);
 
 	_scrollBarWidth = g_gui.xmlEval()->getVar("Globals.Scrollbar.Width", 0);
 
@@ -644,7 +737,7 @@ void ListWidget::reflowLayout() {
 	// of the list.
 	// We do a rough rounding on the decimal places of Entries Per Page,
 	// to add another entry even if it goes a tad over the padding.
-	frac_t entriesPerPage = intToFrac(_h - _topPadding - _bottomPadding) / kLineHeight;
+	frac_t entriesPerPage = intToFrac(_h - _topPadding - _bottomPadding) / (kLineHeight + _itemSpacing);
 
 	// Our threshold before we add another entry is 0.9375 (0xF000 with FRAC_BITS being 16).
 	const frac_t threshold = intToFrac(15) / 16;
@@ -655,25 +748,19 @@ void ListWidget::reflowLayout() {
 	_entriesPerPage = fracToInt(entriesPerPage);
 	assert(_entriesPerPage > 0);
 
-	delete[] _textWidth;
-	_textWidth = new int[_entriesPerPage];
-
-	for (int i = 0; i < _entriesPerPage; i++)
-		_textWidth[i] = 0;
-
 	if (_scrollBar) {
-		_scrollBar->resize(_w - _scrollBarWidth + 1, 0, _scrollBarWidth, _h);
+		_scrollBar->resize(_w - _scrollBarWidth, 0, _scrollBarWidth, _h, false);
 		scrollBarRecalc();
 		scrollToCurrent();
 	}
 }
 
-void ListWidget::setFilter(const String &filter, bool redraw) {
+void ListWidget::setFilter(const Common::U32String &filter, bool redraw) {
 	// FIXME: This method does not deal correctly with edit mode!
 	// Until we fix that, let's make sure it isn't called while editing takes place
 	assert(!_editMode);
 
-	String filt = filter;
+	Common::U32String filt = filter;
 	filt.toLowercase();
 
 	if (_filter == filt) // Filter was not changed
@@ -683,33 +770,37 @@ void ListWidget::setFilter(const String &filter, bool redraw) {
 
 	if (_filter.empty()) {
 		// No filter -> display everything
-		_list = _dataList;
+
+		_list.clear();
+
+		for (uint i = 0; i < _dataList.size(); ++i)
+			_list.push_back(_dataList[i].orig);
+
 		_listIndex.clear();
 	} else {
-		// Restrict the list to everything which contains all words in _filter
-		// as substrings, ignoring case.
+		// Restrict the list to everything which matches all tokens in _filter, ignoring case.
 
-		Common::StringTokenizer tok(_filter);
-		String tmp;
+		Common::U32StringTokenizer tok(_filter);
+		Common::U32String tmp;
 		int n = 0;
 
 		_list.clear();
 		_listIndex.clear();
 
-		for (StringArray::iterator i = _dataList.begin(); i != _dataList.end(); ++i, ++n) {
-			tmp = *i;
+		for (auto i = _dataList.begin(); i != _dataList.end(); ++i, ++n) {
+			tmp = i->clean;
 			tmp.toLowercase();
 			bool matches = true;
 			tok.reset();
 			while (!tok.empty()) {
-				if (!tmp.contains(tok.nextToken())) {
+				if (!_filterMatcher(_filterMatcherArg, n, tmp, tok.nextToken())) {
 					matches = false;
 					break;
 				}
 			}
 
 			if (matches) {
-				_list.push_back(*i);
+				_list.push_back(i->orig);
 				_listIndex.push_back(n);
 			}
 		}
@@ -731,8 +822,145 @@ void ListWidget::setFilter(const String &filter, bool redraw) {
 		// Such a widget could also (optionally) draw a border (or even different
 		// kinds of borders) around the objects it groups; and also a 'title'
 		// (I am borrowing these "ideas" from the NSBox class in Cocoa :).
-		_boss->draw();
+		g_gui.scheduleTopDialogRedraw();
 	}
 }
 
+Common::U32String ListWidget::getThemeColor(byte r, byte g, byte b) {
+	return Common::U32String::format("\001c%02x%02x%02x", r, g, b);
+}
+
+Common::U32String ListWidget::getThemeColor(ThemeEngine::FontColor color) {
+	switch (color) {
+	case ThemeEngine::kFontColorNormal:
+		return Common::U32String("\001C{normal}");
+	case ThemeEngine::kFontColorAlternate:
+		return Common::U32String("\001C{alternate}");
+	default:
+		return Common::U32String("\001C{unknown}");
+	}
+}
+
+ThemeEngine::FontColor ListWidget::getThemeColor(const Common::U32String &color) {
+	if (color == "normal")
+		return ThemeEngine::kFontColorNormal;
+
+	if (color == "alternate")
+		return ThemeEngine::kFontColorAlternate;
+
+	warning("ListWidget::getThemeColor(): Malformed color (\"%s\")", color.encode().c_str());
+
+	return ThemeEngine::kFontColorNormal;
+}
+
+Common::U32String ListWidget::stripGUIformatting(const Common::U32String &str) {
+	Common::U32String stripped;
+	const Common::u32char_type_t *s = str.c_str();
+
+	while (*s) {
+		if (*s != '\001') { // normal symbol
+			stripped += *s++;
+			continue;
+		}
+
+		s++; // skip \001
+		switch (*s) {
+		case '\001':  // \001\001 -> \001
+			stripped += *s++;
+			break;
+
+		case 'c': // \001cRRGGBB
+			s += 7; // check length?
+			break;
+
+		case 'C': // \001C{color-name}
+			while (*s && *s++ != '}')
+				;
+			break;
+
+		default:
+			warning("Wrong string format (%c)", *s ? *s : '?');
+			break;
+		}
+	}
+
+	return stripped;
+}
+
+Common::U32String ListWidget::escapeString(const Common::U32String &str) {
+	Common::U32String escaped;
+	const Common::u32char_type_t *s = str.c_str();
+
+	while (*s) {
+		if (*s == '\001')
+			escaped += '\001';
+		escaped += *s++;
+	}
+
+	return escaped;
+}
+
+void ListWidget::drawFormattedText(const Common::Rect &r, const Common::U32String &str, ThemeEngine::WidgetStateInfo state,
+				Graphics::TextAlign align, ThemeEngine::TextInversionState inverted, int deltax, bool useEllipsis,
+				ThemeEngine::FontColor color) {
+	Common::U32String chunk;
+	const Common::u32char_type_t *s = str.c_str();
+	ThemeEngine::FontStyle curfont = ThemeEngine::kFontStyleBold;
+	ThemeEngine::FontColor curcolor = ThemeEngine::kFontColorNormal;
+	Common::U32String tmp;
+
+	while (*s) {
+		if (*s != '\001') { // normal symbol
+			chunk += *s++;
+			continue;
+		}
+
+		if (chunk.size()) {
+			g_gui.theme()->drawText(r, chunk, state, align, inverted, deltax, true, curfont, curcolor);
+
+			deltax += g_gui.theme()->getStringWidth(chunk, curfont);
+			chunk.clear();
+		}
+
+		s++; // skip \001
+		switch (*s) {
+		case '\001':  // \001\001 -> \001
+			chunk += *s++;
+			break;
+
+		case 'c': // \001cRRGGBB
+			s += 7; // check length?
+			break;
+
+		case 'C': // \001C{color-name}
+			tmp.clear();
+			s++;
+			if (*s == '{')
+				s++;
+			else
+				error("ListWidget::drawFormattedText(): Malformatted \\001C color (%c)", *s);
+
+			while (*s && *s != '}')
+				tmp += *s++;
+
+			if (*s == '}')	// skip the closing bracket
+				s++;
+
+			if (color == ThemeEngine::kFontColorFormatting)
+				curcolor = getThemeColor(tmp);
+			else
+				curcolor = color;	// Ignore color and use the requested one
+
+			break;
+
+		default:
+			warning("ListWidget::drawFormattedText(): Wrong string format (\\001%c)", *s ? *s : '?');
+			break;
+		}
+	}
+
+	if (chunk.size() || str.empty())
+		g_gui.theme()->drawText(r, chunk, state, align, inverted, deltax, true, curfont, curcolor);
+}
+
 } // End of namespace GUI
